# pylint: disable=too-many-instance-attributes,too-many-arguments,too-many-locals

import collections.abc

from .definitions import MatchingRuleKind, AttributeTypeUsage, MatchingRuleUseDefinition, AttributeTypeDefinition, ObjectClassDefinition
from .. import exceptions

__all__ = [
	'Syntax',
	'EqualityMatchingRule',
	'OrderingMatchingRule',
	'SubstrMatchingRule',
	'AttributeType',
	'ObjectClass',
	'Schema',
]

class Syntax:
	'''LDAP syntax for attribute and assertion values'''
	def __init__(self, schema, definition, syntaxes_by_tag):
		self.schema = schema
		self.definition = definition
		self.oid = definition.oid
		self.ref = self.oid
		schema._register(self, self.oid, self.ref)
		schema.syntaxes._register(self, self.oid, self.ref)
		for tag in definition.compatability_tags:
			syntaxes_by_tag.setdefault(tag, set()).add(self)
		# Populated by MatchingRule.__init__
		self.compatible_matching_rules = set()

	def __repr__(self):
		return f'<ldapserver.schema.Syntax {self.oid}>'

	def encode(self, value):
		return self.definition.encode(self.schema, value)

	def decode(self, raw_value):
		return self.definition.decode(self.schema, raw_value)

class MatchingRule:
	def __init__(self, schema, definition, syntaxes_by_tag):
		self.schema = schema
		self.definition = definition
		self.oid = definition.oid
		self.syntax = schema.syntaxes[definition.syntax]
		self.names = self.definition.name
		self.ref = self.names[0] if self.names else self.oid
		schema._register(self, self.oid, self.ref, *self.names)
		schema.matching_rules._register(self, self.oid, self.ref, *self.names)
		self.compatible_syntaxes = syntaxes_by_tag.setdefault(definition.compatability_tag, set())
		for syntax in self.compatible_syntaxes:
			syntax.compatible_matching_rules.add(self)
		# Populated by AttributeType.__init__
		self.compatible_attribute_types = set()

	def match_extensible(self, attribute_values, assertion_value):
		raise exceptions.LDAPInappropriateMatching()

class EqualityMatchingRule(MatchingRule):
	def __repr__(self):
		return f'<ldapserver.schema.EqualityMatchingRule {self.ref}>'

	def match_extensible(self, attribute_values, assertion_value):
		return self.definition.match_equal(self.schema, attribute_values, assertion_value)

	def match_equal(self, attribute_values, assertion_value):
		return self.definition.match_equal(self.schema, attribute_values, assertion_value)

	def match_approx(self, attribute_values, assertion_value):
		return self.definition.match_approx(self.schema, attribute_values, assertion_value)

class OrderingMatchingRule(MatchingRule):
	def __repr__(self):
		return f'<ldapserver.schema.OrderingMatchingRule {self.ref}>'

	def match_less(self, attribute_values, assertion_value):
		return self.definition.match_less(self.schema, attribute_values, assertion_value)

	def match_greater_or_equal(self, attribute_values, assertion_value):
		return self.definition.match_greater_or_equal(self.schema, attribute_values, assertion_value)

class SubstrMatchingRule(MatchingRule):
	def __repr__(self):
		return f'<ldapserver.schema.SubstrMatchingRule {self.ref}>'

	def match_extensible(self, attribute_values, assertion_value):
		return self.definition.match_substr(self.schema, attribute_values, assertion_value[0], assertion_value[1], assertion_value[2])

	def match_substr(self, attribute_values, inital_substring, any_substrings, final_substring):
		return self.definition.match_substr(self.schema, attribute_values, inital_substring, any_substrings, final_substring)

class AttributeType:
	def __init__(self, schema, definition):
		self.schema = schema
		self.definition = definition
		self.oid = definition.oid
		self.names = definition.name or []
		self.ref = self.names[0] if self.names else self.oid
		self.sup = schema.attribute_types[definition.sup] if definition.sup else None
		self.subtypes = set()
		sup = self.sup
		while sup:
			self.sup.subtypes.add(self)
			sup = sup.sup
		self.equality = schema.matching_rules[definition.equality] if definition.equality else None
		self.ordering = schema.matching_rules[definition.ordering] if definition.ordering else None
		self.substr = schema.matching_rules[definition.substr] if definition.substr else None
		if self.sup:
			self.equality = self.equality or self.sup.equality
			self.ordering = self.ordering or self.sup.ordering
			self.substr = self.substr or self.sup.substr
		self.syntax = schema.syntaxes[definition.syntax] if definition.syntax else self.sup.syntax
		self.is_operational = (definition.usage != AttributeTypeUsage.userApplications)
		schema._register(self, self.oid, self.ref, *self.names)
		schema.attribute_types._register(self, self.oid, self.ref, *self.names)
		if not self.is_operational:
			schema.user_attribute_types.add(self)
		self.compatible_matching_rules = self.syntax.compatible_matching_rules
		for matching_rule in self.compatible_matching_rules:
			matching_rule.compatible_attribute_types.add(self)

	def __repr__(self):
		return f'<ldapserver.schema.AttributeType {self.ref}>'

	def encode(self, value):
		return self.syntax.encode(value)

	def decode(self, raw_value):
		return self.syntax.decode(raw_value)

	def match_equal(self, attribute_values, assertion_value):
		'''Return whether any attribute value equals assertion value

		:param attribute_values: Attribute values (type according to syntax)
		:type attribute_values: List of any
		:param assertion_value: Assertion value
		:type assertion_value: bytes
		:returns: True if any attribute values matches, False otherwise
		:rtype: bool
		:raises exceptions.LDAPError: if the result is undefined'''
		if self.equality is None:
			raise exceptions.LDAPInappropriateMatching()
		assertion_value = self.equality.syntax.decode(assertion_value)
		return self.equality.match_equal(attribute_values, assertion_value)

	def match_substr(self, attribute_values, inital_substring, any_substrings, final_substring):
		'''Return whether any attribute value matches a substring assertion

		:param attribute_values: Attribute values (type according to syntax)
		:type attribute_values: List of any
		:param inital_substring: Substring to match the beginning (optional)
		:type inital_substring: bytes or None
		:param any_substrings: List of substrings to match between initial and final in order
		:type any_substrings: list of bytes
		:param final_substring: Substring to match the end (optional)
		:type final_substring: bytes or None
		:returns: True if any attribute values matches, False otherwise
		:rtype: bool
		:raises exceptions.LDAPError: if the result is undefined'''
		if self.equality is None or self.substr is None:
			raise exceptions.LDAPInappropriateMatching()
		if inital_substring:
			inital_substring = self.equality.syntax.decode(inital_substring)
		any_substrings = [self.equality.syntax.decode(substring) for substring in any_substrings]
		if final_substring:
			final_substring = self.equality.syntax.decode(final_substring)
		return self.substr.match_substr(attribute_values, inital_substring, any_substrings, final_substring)

	def match_approx(self, attribute_values, assertion_value):
		'''Return whether any attribute value approximatly equals assertion value

		:param attribute_values: Attribute values (type according to syntax)
		:type attribute_values: List of any
		:param assertion_value: Assertion value
		:type assertion_value: bytes
		:returns: True if any attribute values matches, False otherwise
		:rtype: bool
		:raises exceptions.LDAPError: if the result is undefined'''
		if self.equality is None:
			raise exceptions.LDAPInappropriateMatching()
		assertion_value = self.equality.syntax.decode(assertion_value)
		return self.equality.match_approx(attribute_values, assertion_value)

	def match_greater_or_equal(self, attribute_values, assertion_value):
		if self.ordering is None:
			raise exceptions.LDAPInappropriateMatching()
		assertion_value = self.ordering.syntax.decode(assertion_value)
		return self.ordering.match_greater_or_equal(attribute_values, assertion_value)

	def __match_less(self, attribute_values, assertion_value):
		if self.ordering is None:
			raise exceptions.LDAPInappropriateMatching()
		assertion_value = self.ordering.syntax.decode(assertion_value)
		return self.ordering.match_less(attribute_values, assertion_value)

	def match_less_or_equal(self, attribute_values, assertion_value):
		equal_exc = None
		try:
			if self.match_equal(attribute_values, assertion_value):
				return True
		except exceptions.LDAPError as exc:
			equal_exc = exc
		if self.__match_less(attribute_values, assertion_value):
			return True
		if equal_exc is not None:
			raise equal_exc
		return False

	def match_extensible(self, attribute_values, assertion_value, matching_rule=None):
		if not matching_rule:
			matching_rule = self.equality
		if not matching_rule or matching_rule not in self.compatible_matching_rules:
			raise exceptions.LDAPInappropriateMatching()
		assertion_value = matching_rule.syntax.decode(assertion_value)
		return matching_rule.match_extensible(attribute_values, assertion_value)

class ObjectClass:
	'''Representation of an object class wihin a schema'''
	def __init__(self, schema, definition):
		self.schema = schema
		self.definition = definition
		self.oid = definition.oid
		self.names = definition.name
		self.ref = self.names[0] if self.names else self.oid
		# Lookup dependencies to ensure consistency
		# pylint: disable=pointless-statement
		for sup_oid in definition.sup:
			schema.object_classes[sup_oid]
		for must_oid in definition.must:
			schema.attribute_types[must_oid]
		for may_oid in definition.may:
			schema.attribute_types[may_oid]
		schema._register(self, self.oid, self.ref, *self.names)
		schema.object_classes._register(self, self.oid, self.ref, *self.names)

	def __repr__(self):
		return f'<ldapserver.schema.ObjectClass {self.ref}>'

class OIDDict(collections.abc.Mapping):
	def __init__(self):
		self.__data = {}
		self.__refs = []
		self.__numeric_oid_map = {}

	def _register(self, obj, oid, ref, *names):
		if obj in self.__data:
			return
		if oid in self.__data:
			raise Exception(f'OID "{oid}" already registered')
		for name in (ref,) + names:
			if name.lower().strip() in self.__data:
				raise Exception(f'Short descriptive name "{name}" already registered')
		self.__refs.append(ref)
		self.__data[obj] = obj
		self.__numeric_oid_map[obj] = oid
		for name in (oid, ref) + names:
			self.__data[name.lower().strip()] = obj
			self.__numeric_oid_map[name.lower().strip()] = oid

	def __getitem__(self, key):
		if isinstance(key, str):
			key = key.lower().strip()
		return self.__data[key]

	def __iter__(self):
		return iter(self.__refs)

	def __len__(self):
		return len(self.__refs)

	def get_numeric_oid(self, key, default=None):
		if isinstance(key, str):
			key = key.lower().strip()
		return self.__numeric_oid_map.get(key, default)

class Schema(OIDDict):
	'''Collection of LDAP syntaxes, matching rules, attribute types and object
	classes forming an LDAP schema.'''
	# pylint: disable=too-many-branches,too-many-statements
	def __init__(self, object_class_definitions=None, attribute_type_definitions=None,
	             matching_rule_definitions=None, syntax_definitions=None):
		super().__init__()
		syntaxes_by_tag = {}

		# Add syntaxes
		self.syntaxes = OIDDict()
		for definition in syntax_definitions or []:
			if definition.oid not in self.syntaxes:
				Syntax(self, definition, syntaxes_by_tag)
		self.syntax_definitions = [syntax.definition for syntax in self.syntaxes.values()]

		# Add matching rules
		self.matching_rules = OIDDict()
		for definition in matching_rule_definitions or []:
			if definition.kind == MatchingRuleKind.EQUALITY:
				cls = EqualityMatchingRule
			elif definition.kind == MatchingRuleKind.ORDERING:
				cls = OrderingMatchingRule
			elif definition.kind == MatchingRuleKind.SUBSTR:
				cls = SubstrMatchingRule
			else:
				raise ValueError('Invalid matching rule kind')
			if definition.oid not in self.matching_rules:
				cls(self, definition, syntaxes_by_tag)
		self.matching_rule_definitions = [matching_rule.definition for matching_rule in self.matching_rules.values()]

		# Add attribute types
		attribute_type_definitions = [AttributeTypeDefinition.from_str(item)
		                              if isinstance(item, str) else item
		                              for item in attribute_type_definitions or []]
		self.attribute_types = OIDDict()
		self.user_attribute_types = set()
		# Attribute types may refer to other (superior) attribute types. To resolve
		# these dependencies we cycle through the definitions, each time adding
		# those not added yet with fulfilled dependencies. Finally we add all the
		# remaining ones to provoke exceptions.
		keep_running = True
		while keep_running:
			keep_running = False
			for definition in attribute_type_definitions:
				if definition.oid in self.attribute_types:
					continue
				if definition.sup and definition.sup not in self.attribute_types:
					continue
				AttributeType(self, definition)
				keep_running = True
		for definition in attribute_type_definitions:
			if definition.oid not in self.attribute_types:
				AttributeType(self, definition)
		self.attribute_type_definitions = [attribute_type.definition for attribute_type in self.attribute_types.values()]

		# Add object classes
		object_class_definitions = [ObjectClassDefinition.from_str(item)
		                            if isinstance(item, str) else item
		                            for item in object_class_definitions or []]
		self.object_classes = OIDDict()
		# Object classes may refer to other (superior) object classes. To resolve
		# these dependencies we cycle through the definitions, each time adding
		# those not added yet with fulfilled dependencies. Finally we add all the
		# remaining ones to provoke exceptions.
		keep_running = True
		while keep_running:
			keep_running = False
			for definition in object_class_definitions:
				if definition.oid in self.object_classes:
					continue
				if any(map(lambda oid: oid and oid not in self.object_classes, definition.sup)):
					continue
				ObjectClass(self, definition)
				keep_running = True
		for definition in object_class_definitions:
			if definition.oid not in self.object_classes:
				ObjectClass(self, definition)
		self.object_class_definitions = [object_class.definition for object_class in self.object_classes.values()]

		# Generate and add matching rules
		self.matching_rule_use_definitions = []
		for matching_rule in self.matching_rules.values():
			definition = matching_rule.definition
			applies = [attribute_type.ref for attribute_type in matching_rule.compatible_attribute_types]
			rule_use = MatchingRuleUseDefinition(definition.oid, name=definition.name,
					desc=definition.desc, obsolete=definition.obsolete, applies=applies,
					extensions=definition.extensions)
			if applies:
				self.matching_rule_use_definitions.append(rule_use)

	def extend(self, object_class_definitions=None, attribute_type_definitions=None,
	           matching_rule_definitions=None, syntax_definitions=None):
		syntax_definitions = self.syntax_definitions + (syntax_definitions or [])
		matching_rule_definitions = self.matching_rule_definitions + (matching_rule_definitions or [])
		attribute_type_definitions = self.attribute_type_definitions + (attribute_type_definitions or [])
		object_class_definitions = self.object_class_definitions + (object_class_definitions or [])
		return type(self)(syntax_definitions=syntax_definitions,
		                  matching_rule_definitions=matching_rule_definitions,
		                  attribute_type_definitions=attribute_type_definitions,
		                  object_class_definitions=object_class_definitions)

	def __or__(self, value):
		return self.extend(syntax_definitions=value.syntax_definitions,
		                   matching_rule_definitions=value.matching_rule_definitions,
		                   attribute_type_definitions=value.attribute_type_definitions,
		                   object_class_definitions=value.object_class_definitions)



import collections.abc
import enum

from . import ldap, exceptions
from .dn import DN, RDN, RDNAssertion

class TypeKeysView(collections.abc.Set):
	def __init__(self, attributes):
		self.__attributes = attributes

	def __iter__(self):
		for attribute_type, values in self.__attributes.items():
			if values:
				yield attribute_type

	def __len__(self):
		return len(list(iter(self)))

	def __contains__(self, value):
		return bool(self.__attributes.get(value))

class TypeItemsView(collections.abc.Set):
	def __init__(self, attributes):
		self.__attributes = attributes

	def __iter__(self):
		for attribute_type, values in self.__attributes.items():
			if values:
				yield attribute_type, values

	def __len__(self):
		return len(list(iter(self)))

	def __contains__(self, value):
		key, values = value
		return self.__attributes.get(key) == values

class AttributeDict(collections.abc.MutableMapping):
	'''Special dictionary holding LDAP attribute values

	Attribute values can be set and accessed by their attribute type's numeric
	OID or short descriptive name. Attribute types must be defined within the
	schema to be used. Attribute values are always lists. Accessing an unset
	attribute behaves the same as accessing an empty attribute. List items must
	conform to the attribute's syntax.'''
	def __init__(self, schema, **attributes):
		self.__attributes = {}
		self.schema = schema
		for key, values in attributes.items():
			self[key] = values

	def __getitem__(self, key):
		return self.__attributes.setdefault(self.schema.attribute_types[key], [])

	def __setitem__(self, key, values):
		self.__attributes[self.schema.attribute_types[key]] = values

	def __delitem__(self, key):
		self[key] = []

	def __iter__(self):
		for attribute_type, values in self.__attributes.items():
			if values:
				yield attribute_type.ref

	def __len__(self):
		return len(list(iter(self)))

	def get(self, key, default=None, subtypes=False): # pylint: disable=arguments-differ
		attribute_type = self.schema.attribute_types.get(key)
		attribute_types = [attribute_type]
		if subtypes and attribute_type:
			attribute_types += attribute_type.subtypes
		result = []
		for attribute_type in attribute_types:
			result += self.__attributes.get(attribute_type, [])
		if not result:
			result = default if default is not None else []
		return result

	def keys(self, types=False): # pylint: disable=arguments-differ
		if not types:
			return super().keys()
		return TypeKeysView(self.__attributes)

	def items(self, types=False): # pylint: disable=arguments-differ
		if not types:
			return super().items()
		return TypeItemsView(self.__attributes)

	def __contains__(self, key):
		try:
			return bool(self[key])
		except KeyError:
			return False

	def setdefault(self, key, default=False):
		if key in self:
			return self[key]
		if default is None:
			default = []
		self[key] = default
		return default

class FilterResult(enum.Enum):
	TRUE = enum.auto()
	FALSE = enum.auto()
	UNDEFINED = enum.auto()

class Object(AttributeDict):
	def __init__(self, schema, dn, **attributes):
		super().__init__(schema, **attributes)
		self.dn = DN(schema, dn)

	def __search_match_dn(self, basedn, scope):
		if scope == ldap.SearchScope.baseObject:
			return self.dn == basedn
		elif scope == ldap.SearchScope.singleLevel:
			return self.dn.is_direct_child_of(basedn)
		elif scope == ldap.SearchScope.wholeSubtree:
			return self.dn.in_subtree_of(basedn)
		else:
			return False

	def __search_match_filter(self, filter_obj):
		# pylint: disable=too-many-branches,too-many-return-statements,too-many-statements,too-many-nested-blocks
		if isinstance(filter_obj, ldap.FilterAnd):
			result = FilterResult.TRUE
			for subfilter in filter_obj.filters:
				subresult = self.__search_match_filter(subfilter)
				if subresult == FilterResult.FALSE:
					return FilterResult.FALSE
				elif subresult == FilterResult.UNDEFINED:
					result = FilterResult.UNDEFINED
			return result
		elif isinstance(filter_obj, ldap.FilterOr):
			result = FilterResult.FALSE
			for subfilter in filter_obj.filters:
				subresult = self.__search_match_filter(subfilter)
				if subresult == FilterResult.TRUE:
					return FilterResult.TRUE
				elif subresult == FilterResult.UNDEFINED:
					result = FilterResult.UNDEFINED
			return result
		elif isinstance(filter_obj, ldap.FilterNot):
			subresult = self.__search_match_filter(filter_obj.filter)
			if subresult == FilterResult.TRUE:
				return FilterResult.FALSE
			elif subresult == FilterResult.FALSE:
				return FilterResult.TRUE
			else:
				return subresult
		elif isinstance(filter_obj, (ldap.FilterPresent,
		                             ldap.FilterEqual,
		                             ldap.FilterSubstrings,
		                             ldap.FilterApproxMatch,
		                             ldap.FilterGreaterOrEqual,
		                             ldap.FilterLessOrEqual)):
			try:
				attribute_type = self.schema.attribute_types[filter_obj.attribute]
			except KeyError:
				return FilterResult.UNDEFINED
			values = self.get(filter_obj.attribute, subtypes=True)
			try:
				if isinstance(filter_obj, ldap.FilterPresent):
					result = values != []
				elif isinstance(filter_obj, ldap.FilterEqual):
					result = attribute_type.match_equal(values, filter_obj.value)
				elif isinstance(filter_obj, ldap.FilterSubstrings):
					result = attribute_type.match_substr(values,
							filter_obj.initial_substring, filter_obj.any_substrings, filter_obj.final_substring)
				elif isinstance(filter_obj, ldap.FilterApproxMatch):
					result = attribute_type.match_approx(values, filter_obj.value)
				elif isinstance(filter_obj, ldap.FilterGreaterOrEqual):
					result = attribute_type.match_greater_or_equal(values, filter_obj.value)
				elif isinstance(filter_obj, ldap.FilterLessOrEqual):
					result = attribute_type.match_less_or_equal(values, filter_obj.value)
				else:
					return FilterResult.UNDEFINED
				return FilterResult.TRUE if result else FilterResult.FALSE
			except exceptions.LDAPError:
				return FilterResult.UNDEFINED
		elif isinstance(filter_obj, ldap.FilterExtensibleMatch):
			attribute_types = []
			matching_rule = None
			try:
				if filter_obj.type is not None and filter_obj.matchingRule is not None:
					attribute_types = [self.schema.attribute_types[filter_obj.type]]
					matching_rule = self.schema.matching_rules[filter_obj.matchingRule]
				elif filter_obj.type is not None:
					attribute_types = [self.schema.attribute_types[filter_obj.type]]
				elif filter_obj.matchingRule is not None:
					matching_rule = self.schema.matching_rules[filter_obj.matchingRule]
					attribute_types = matching_rule.compatible_attribute_types
			except KeyError:
				pass
			result = FilterResult.FALSE
			for attribute_type in attribute_types:
				values = self.get(attribute_type.oid, subtypes=True)
				if filter_obj.dnAttributes:
					for rdn in self.dn:
						for assertion in rdn:
							if assertion.attribute.lower() == attribute_type.ref.lower():
								values.append(assertion.value)
				try:
					if attribute_type.match_extensible(values, filter_obj.matchValue, matching_rule):
						return FilterResult.TRUE
				except exceptions.LDAPError:
					result = FilterResult.UNDEFINED
			return result
		else:
			return FilterResult.UNDEFINED

	def match_search(self, base_obj, scope, filter_obj):
		return self.__search_match_dn(DN.from_str(self.schema, base_obj), scope) and \
		       self.__search_match_filter(filter_obj) == FilterResult.TRUE

	def search(self, base_obj, scope, filter_obj, attributes, types_only):
		if not self.match_search(base_obj, scope, filter_obj):
			return None
		selected_attributes = set()
		for selector in attributes or ['*']:
			if selector == '*':
				selected_attributes |= self.schema.user_attribute_types
			elif selector == '1.1':
				continue
			elif selector in self.schema.attribute_types:
				selected_attributes.add(self.schema.attribute_types[selector])
		partial_attributes = []
		for attribute_type, values in self.items(types=True):
			if attribute_type in selected_attributes:
				if types_only:
					values = []
				encoded_values = [attribute_type.encode(value) for value in values]
				partial_attributes.append(ldap.PartialAttribute(attribute_type.ref, encoded_values))
		return ldap.SearchResultEntry(str(self.dn), partial_attributes)

	def compare(self, dn, attribute, value):
		try:
			dn = DN.from_str(self.schema, dn)
		except ValueError as exc:
			raise exceptions.LDAPNoSuchObject() from exc
		if dn != self.dn:
			raise exceptions.LDAPNoSuchObject()
		try:
			attribute_type = self.schema.attribute_types[attribute]
		except KeyError as exc:
			raise exceptions.LDAPUndefinedAttributeType() from exc
		return attribute_type.match_equal(self.get(attribute_type, subtypes=True), value)

class RootDSE(Object):
	def __init__(self, schema, *args, **kwargs):
		super().__init__(schema, DN(schema), *args, **kwargs)
		self.setdefault('objectClass', ['top'])

	def match_search(self, base_obj, scope, filter_obj):
		return not DN.from_str(self.schema, base_obj) and scope == ldap.SearchScope.baseObject and \
		       isinstance(filter_obj, ldap.FilterPresent) and \
		       filter_obj.attribute.lower() == 'objectclass'

class WildcardValue:
	pass

WILDCARD_VALUE = WildcardValue()

class TemplateFilterResult(enum.Enum):
	TRUE = enum.auto()
	FALSE = enum.auto()
	UNDEFINED = enum.auto()
	MAYBE_TRUE = enum.auto()

class ObjectTemplate(AttributeDict):
	def __init__(self, schema, parent_dn, rdn_attribute, **attributes):
		super().__init__(schema, **attributes)
		self.parent_dn = DN(schema, parent_dn)
		self.rdn_attribute = rdn_attribute

	def __match_extract_dn_constraints(self, basedn, scope):
		if scope == ldap.SearchScope.baseObject:
			if basedn[1:] != self.parent_dn or basedn.object_attribute.lower() != self.rdn_attribute.lower():
				return False, AttributeDict(self.schema)
			return True, AttributeDict(self.schema, **{self.rdn_attribute: [basedn.object_value]})
		elif scope == ldap.SearchScope.singleLevel:
			return basedn == self.parent_dn, AttributeDict(self.schema)
		elif scope == ldap.SearchScope.wholeSubtree:
			if self.parent_dn.in_subtree_of(basedn):
				return True, AttributeDict(self.schema)
			if basedn[1:] != self.parent_dn or basedn.object_attribute.lower() != self.rdn_attribute.lower():
				return False, AttributeDict(self.schema)
			return True, AttributeDict(self.schema, **{self.rdn_attribute: [basedn.object_value]})
		else:
			return False, AttributeDict(self.schema)

	def __search_match_dn(self, basedn, scope):
		'''Return whether objects from this template might match the provided parameters'''
		return self.__match_extract_dn_constraints(basedn, scope)[0]

	def __extract_dn_constraints(self, basedn, scope):
		return self.__match_extract_dn_constraints(basedn, scope)[1]

	def __search_match_filter(self, filter_obj):
		# pylint: disable=too-many-return-statements,too-many-branches,too-many-nested-blocks,too-many-statements
		if isinstance(filter_obj, ldap.FilterAnd):
			result = TemplateFilterResult.TRUE
			for subfilter in filter_obj.filters:
				subresult = self.__search_match_filter(subfilter)
				if subresult == TemplateFilterResult.FALSE:
					return TemplateFilterResult.FALSE
				elif subresult == TemplateFilterResult.UNDEFINED:
					result = TemplateFilterResult.UNDEFINED
				elif subresult == TemplateFilterResult.MAYBE_TRUE and result == TemplateFilterResult.TRUE:
					result = TemplateFilterResult.MAYBE_TRUE
			return result
		elif isinstance(filter_obj, ldap.FilterOr):
			result = TemplateFilterResult.FALSE
			for subfilter in filter_obj.filters:
				subresult = self.__search_match_filter(subfilter)
				if subresult == TemplateFilterResult.TRUE:
					return TemplateFilterResult.TRUE
				elif subresult == TemplateFilterResult.MAYBE_TRUE:
					result = TemplateFilterResult.MAYBE_TRUE
				elif subresult == TemplateFilterResult.UNDEFINED and result == TemplateFilterResult.FALSE:
					result = TemplateFilterResult.UNDEFINED
			return result
		elif isinstance(filter_obj, ldap.FilterNot):
			subresult = self.__search_match_filter(filter_obj.filter)
			if subresult == TemplateFilterResult.TRUE:
				return TemplateFilterResult.FALSE
			elif subresult == TemplateFilterResult.FALSE:
				return TemplateFilterResult.TRUE
			else:
				return subresult
		elif isinstance(filter_obj, (ldap.FilterPresent,
		                             ldap.FilterEqual,
		                             ldap.FilterSubstrings,
		                             ldap.FilterApproxMatch,
		                             ldap.FilterGreaterOrEqual,
		                             ldap.FilterLessOrEqual)):
			try:
				attribute_type = self.schema.attribute_types[filter_obj.attribute]
			except KeyError:
				return TemplateFilterResult.UNDEFINED
			values = self.get(filter_obj.attribute, subtypes=True)
			is_wildcard = WILDCARD_VALUE in values
			if is_wildcard:
				values = []
			try:
				if isinstance(filter_obj, ldap.FilterPresent):
					result = values != []
				elif isinstance(filter_obj, ldap.FilterEqual):
					result = attribute_type.match_equal(values, filter_obj.value)
				elif isinstance(filter_obj, ldap.FilterSubstrings):
					result = attribute_type.match_substr(values,
							filter_obj.initial_substring, filter_obj.any_substrings, filter_obj.final_substring)
				elif isinstance(filter_obj, ldap.FilterApproxMatch):
					result = attribute_type.match_approx(values, filter_obj.value)
				elif isinstance(filter_obj, ldap.FilterGreaterOrEqual):
					result = attribute_type.match_greater_or_equal(values, filter_obj.value)
				elif isinstance(filter_obj, ldap.FilterLessOrEqual):
					result = attribute_type.match_less_or_equal(values, filter_obj.value)
				else:
					return TemplateFilterResult.UNDEFINED
				if result:
					return TemplateFilterResult.TRUE
			except exceptions.LDAPError:
				return TemplateFilterResult.UNDEFINED
			if is_wildcard:
				return TemplateFilterResult.MAYBE_TRUE
			return TemplateFilterResult.FALSE
		elif isinstance(filter_obj, ldap.FilterExtensibleMatch):
			attribute_types = []
			matching_rule = None
			try:
				if filter_obj.type is not None and filter_obj.matchingRule is not None:
					attribute_types = [self.schema.attribute_types[filter_obj.type]]
					matching_rule = self.schema.matching_rules[filter_obj.matchingRule]
				elif filter_obj.type is not None:
					attribute_types = [self.schema.attribute_types[filter_obj.type]]
				elif filter_obj.matchingRule is not None:
					matching_rule = self.schema.matching_rules[filter_obj.matchingRule]
					attribute_types = matching_rule.compatible_attribute_types
			except KeyError:
				pass
			result = TemplateFilterResult.FALSE
			for attribute_type in attribute_types:
				values = self.get(attribute_type.oid, subtypes=True)
				if filter_obj.dnAttributes:
					for rdn in self.parent_dn:
						for assertion in rdn:
							if assertion.attribute.lower() == attribute_type.ref.lower():
								values.append(assertion.value)
				is_wildcard = WILDCARD_VALUE in values
				if is_wildcard:
					values = []
				is_undefined = False
				try:
					if attribute_type.match_extensible(values, filter_obj.matchValue, matching_rule):
						return TemplateFilterResult.TRUE
				except exceptions.LDAPError:
					is_undefined = True
				if is_undefined and result == TemplateFilterResult.FALSE:
					result = TemplateFilterResult.UNDEFINED
				elif not is_undefined and is_wildcard:
					result = TemplateFilterResult.MAYBE_TRUE
			return result
		else:
			return TemplateFilterResult.UNDEFINED

	def __extract_filter_constraints(self, filter_obj):
		if isinstance(filter_obj, ldap.FilterEqual):
			try:
				attribute_type = self.schema.attribute_types[filter_obj.attribute]
			except KeyError:
				return AttributeDict(self.schema)
			if attribute_type.equality is None:
				return AttributeDict(self.schema)
			assertion_value = attribute_type.equality.syntax.decode(filter_obj.value)
			if assertion_value is None:
				return AttributeDict(self.schema)
			return AttributeDict(self.schema, **{filter_obj.attribute: [assertion_value]})
		if isinstance(filter_obj, ldap.FilterAnd):
			result = AttributeDict(self.schema)
			for subfilter in filter_obj.filters:
				for name, values in self.__extract_filter_constraints(subfilter).items():
					result[name] += values
			return result
		return AttributeDict(self.schema)

	def match_search(self, base_obj, scope, filter_obj):
		'''Return whether objects based on this template might match the search parameters'''
		return self.__search_match_dn(DN.from_str(self.schema, base_obj), scope) and \
		       self.__search_match_filter(filter_obj) in (TemplateFilterResult.TRUE,
		                                                  TemplateFilterResult.MAYBE_TRUE)

	def extract_search_constraints(self, base_obj, scope, filter_obj):
		constraints = self.__extract_filter_constraints(filter_obj)
		for key, values in self.__extract_dn_constraints(DN.from_str(self.schema, base_obj), scope).items():
			constraints[key] += values
		return constraints

	def create_object(self, rdn_value, **attributes):
		obj = Object(self.schema, DN(self.schema, self.parent_dn, **{self.rdn_attribute: rdn_value}))
		for key, values in attributes.items():
			if WILDCARD_VALUE not in self[key]:
				raise ValueError(f'Cannot set attribute "{key}" that is not set to [WILDCARD_VALUE] in the template')
			obj[key] = values
		for attribute_type, values in self.items():
			if WILDCARD_VALUE not in values:
				obj[attribute_type] = values
		return obj

class SubschemaSubentry(Object):
	'''Special :any:`Object` providing information on a Schema'''
	def __init__(self, schema, dn, **attributes):
		super().__init__(schema, dn, **attributes)
		self['subschemaSubentry'] = [self.dn]
		self['structuralObjectClass'] = ['subtree']
		self['objectClass'] = ['top', 'subtree', 'subschema']
		self['objectClasses'] = schema.object_class_definitions
		self['ldapSyntaxes'] = schema.syntax_definitions
		self['matchingRules'] = schema.matching_rule_definitions
		self['attributeTypes'] = schema.attribute_type_definitions
		self['matchingRuleUse'] = schema.matching_rule_use_definitions
		# pylint: disable=invalid-name
		self.AttributeDict = lambda **attributes: AttributeDict(schema, **attributes)
		self.Object = lambda *args, **attributes: Object(schema, *args, subschemaSubentry=[self.dn], **attributes)
		self.RootDSE = lambda **attributes: RootDSE(schema, subschemaSubentry=[self.dn], **attributes)
		self.ObjectTemplate = lambda *args, **kwargs: ObjectTemplate(schema, *args, subschemaSubentry=[self.dn], **kwargs)
		class Wrapper:
			def __init__(self, cls, schema):
				self.cls = cls
				self.schema = schema

			def __call__(self, *args, **kwargs):
				return self.cls(self.schema, *args, **kwargs)

			def from_str(self, *args, **kwargs):
				return self.cls.from_str(self.schema, *args, **kwargs)

		self.DN = Wrapper(DN, schema)
		self.RDN = Wrapper(RDN, schema)
		self.RDNAssertion = Wrapper(RDNAssertion, schema)

	def match_search(self, base_obj, scope, filter_obj):
		return DN.from_str(self.schema, base_obj) == self.dn and  \
		       scope == ldap.SearchScope.baseObject and \
		       isinstance(filter_obj, ldap.FilterEqual) and \
		       filter_obj.attribute.lower() == 'objectclass' and \
		       filter_obj.value.lower() == b'subschema'



'''LDAP Distinguished Name Utilities

Distinguished Names (DNs) identifiy objects in an LDAP directory. In LDAP
protocol messages and when stored in attribute values DNs are encoded with
a string representation scheme described in RFC4514.

This module provides classes to represent `DN` objects and theirs parts
(`RD` and `RDNAssertion`) that correctly implement encoding, decoding and
comparing.

Limitations:

* Hexstring attribute values (`foo=#ABCDEF...`) are not supported
* Attribute values are only validated in from_str
'''

import typing
import re

from . import exceptions

__all__ = ['DN', 'RDN', 'RDNAssertion']

class DN(tuple):
	'''Distinguished Name consiting of zero ore more `RDN` objects'''
	schema: typing.Any

	def __new__(cls, schema, *args, **kwargs):
		if len(args) == 1 and isinstance(args[0], DN):
			args = args[0]
		if len(args) == 1 and isinstance(args[0], str):
			args = cls.from_str(schema, args[0])
		for rdn in args:
			if not isinstance(rdn, RDN):
				raise TypeError(f'Argument {repr(rdn)} is of type {repr(type(rdn))}, expected ldapserver.dn.RDN object')
		rdns = tuple(args)
		if kwargs:
			rdns = (RDN(schema, **kwargs),) + rdns
		dn = super().__new__(cls, rdns)
		dn.schema = schema
		return dn

	# Syntax definiton from RFC4514:
	# distinguishedName = [ relativeDistinguishedName *( COMMA relativeDistinguishedName ) ]

	@classmethod
	def from_str(cls, schema, expr):
		escaped = False
		rdns = []
		token = ''
		for char in expr:
			if escaped:
				escaped = False
				token += char
			elif char == ',':
				rdns.append(RDN.from_str(schema, token))
				token = ''
			else:
				if char == '\\':
					escaped = True
				token += char
		if token:
			rdns.append(RDN.from_str(schema, token))
		return cls(schema, *rdns)

	def __str__(self):
		return ','.join(map(str, self))

	def __repr__(self):
		return '<ldapserver.dn.DN %s>'%str(self)

	def __eq__(self, obj):
		return type(self) is type(obj) and super().__eq__(obj)

	def __ne__(self, obj):
		return not self == obj

	def __add__(self, value):
		if isinstance(value, DN):
			return DN(self.schema, *(tuple(self) + tuple(value)))
		elif isinstance(value, RDN):
			return self + DN(self.schema, value)
		else:
			raise TypeError(f'Can only add DN or RDN to DN, not {type(value)}')

	def __getitem__(self, key):
		if isinstance(key, slice):
			return type(self)(self.schema, *super().__getitem__(key))
		return super().__getitem__(key)

	def __strip_common_suffix(self, value):
		value = DN(self.schema, value)
		minlen = min(len(self), len(value))
		for i in range(minlen):
			if self[-1 - i] != value[-1 - i]:
				return self[:-i or None], value[:-i or None]
		return self[:-minlen or None], value[:-minlen or None]

	def is_direct_child_of(self, base):
		rchild, rbase = self.__strip_common_suffix(DN(self.schema, base))
		return not rbase and len(rchild) == 1

	def in_subtree_of(self, base):
		rchild, rbase = self.__strip_common_suffix(DN(self.schema, base)) # pylint: disable=unused-variable
		return not rbase

	@property
	def object_attribute(self):
		if len(self) == 0:
			return None
		return self[0].attribute # pylint: disable=no-member

	@property
	def object_attribute_type(self):
		if len(self) == 0:
			return None
		return self[0].attribute_type # pylint: disable=no-member

	@property
	def object_value(self):
		if len(self) == 0:
			return None
		return self[0].value # pylint: disable=no-member

	@property
	def object_value_normalized(self):
		if len(self) == 0:
			return None
		return self[0].value_normalized # pylint: disable=no-member

class DNWithUID(DN):
	# pylint: disable=arguments-differ,no-member
	def __new__(cls, schema, dn, uid=None):
		if not uid:
			return dn
		if not re.fullmatch(r"'[01]*'B", uid):
			raise ValueError('Invalid uid value')
		obj = super().__new__(cls, schema, *dn)
		obj.uid = uid
		return obj

	@classmethod
	def from_str(cls, schema, expr):
		dn_part, uid_part = (expr.rsplit('#', 1) + [''])[:2]
		return cls(schema, DN.from_str(schema, dn_part), uid_part or None)

	def __str__(self):
		return super().__str__() + '#' + self.uid

	def __repr__(self):
		return '<ldapserver.dn.DNWithUID %s>'%str(self)

	def __eq__(self, obj):
		return type(self) is type(obj) and super().__eq__(obj) and self.uid == obj.uid

	@property
	def dn(self):
		return DN(self.schema, *self)

class RDN(tuple):
	'''Relative Distinguished Name consisting of one or more `RDNAssertion` objects'''
	schema: typing.Any

	def __new__(cls, schema, *assertions, **kwargs):
		for assertion in assertions:
			if not isinstance(assertion, RDNAssertion):
				raise TypeError(f'Argument {repr(assertion)} is of type {repr(type(assertion))}, expected ldapserver.dn.RDNAssertion')
			if assertion.attribute_type.schema is not schema:
				raise ValueError('RDNAssertion has different schema')
		assertions = list(assertions)
		for key, value in kwargs.items():
			assertions.append(RDNAssertion(schema, key, value))
		if not assertions:
			raise ValueError('RDN must have at least one assertion')
		rdn = super().__new__(cls, assertions)
		rdn.schema = schema
		return rdn

	# Syntax definiton from RFC4514:
	# relativeDistinguishedName = attributeTypeAndValue *( PLUS attributeTypeAndValue )

	@classmethod
	def from_str(cls, schema, expr):
		escaped = False
		assertions = []
		token = ''
		for char in expr:
			if escaped:
				escaped = False
				token += char
			elif char == '+':
				assertions.append(RDNAssertion.from_str(schema, token))
				token = ''
			else:
				if char == '\\':
					escaped = True
				token += char
		if token:
			assertions.append(RDNAssertion.from_str(schema, token))
		return cls(schema, *assertions)

	def __str__(self):
		return '+'.join(map(str, self))

	def __repr__(self):
		return '<ldapserver.dn.RDN %s>'%str(self)

	def __eq__(self, obj):
		return type(self) is type(obj) and set(self) == set(obj)

	def __ne__(self, obj):
		return not self == obj

	def __add__(self, value):
		if isinstance(value, RDN):
			return DN(self.schema, self, value)
		elif isinstance(value, DN):
			return DN(self.schema, self) + value
		else:
			raise TypeError(f'Can only add DN or RDN to RDN, not {type(value)}')

	@property
	def attribute(self):
		if len(self) != 1:
			return None
		return self[0].attribute

	@property
	def attribute_type(self):
		if len(self) != 1:
			return None
		return self[0].attribute_type

	@property
	def value(self):
		if len(self) != 1:
			return None
		return self[0].value

DN_ESCAPED = (
	0x0022, # '"'
	0x002B, # '+'
	0x002C, # ','
	0x003B, # ';'
	0x003C, # '<'
	0x003E, # '>'
)
DN_SPECIAL = DN_ESCAPED + (
	0x0020, # ' '
	0x0023, # '#'
	0x003D, # '='
)
HEXDIGITS = (
	0x0030, # '0'
	0x0031, # '1'
	0x0032, # '2'
	0x0033, # '3'
	0x0034, # '4'
	0x0035, # '5'
	0x0036, # '6'
	0x0037, # '7'
	0x0038, # '8'
	0x0039, # '9'
	0x0041, # 'A'
	0x0042, # 'B'
	0x0043, # 'C'
	0x0044, # 'D'
	0x0045, # 'E'
	0x0046, # 'F'
	0x0061, # 'a'
	0x0062, # 'b'
	0x0063, # 'c'
	0x0064, # 'd'
	0x0065, # 'e'
	0x0066, # 'f'
)

class RDNAssertion:
	'''A single attribute value assertion'''
	__slots__ = ['attribute', 'attribute_type', 'value']
	attribute: str
	attribute_type: typing.Any
	value: typing.Any
	schema: typing.Any

	def __init__(self, schema, attribute, value):
		try:
			super().__setattr__('attribute_type', schema.attribute_types[attribute])
		except KeyError as exc:
			raise ValueError('Invalid RDN attribute type: Attribute type undefined in schema') from exc
		super().__setattr__('attribute', attribute)
		if not self.attribute_type.equality:
			raise ValueError('Invalid RDN attribute type: Attribute type has no EQUALITY matching rule')
		super().__setattr__('value', value)

	# Syntax definiton from RFC4514 and 4512:
	#
	# attributeTypeAndValue = attributeType EQUALS attributeValue
	# attributeType = descr / numericoid
	# attributeValue = string / hexstring
	#
	# descr = ALPHA *( ALPHA / DIGIT / HYPHEN )
	# numericoid = number 1*( DOT number )
	# number = DIGIT / ( LDIGIT 1*DIGIT )
	#
	# ; The following characters are to be escaped when they appear
	# ; in the value to be encoded: ESC, one of <escaped>, leading
	# ; SHARP or SPACE, trailing SPACE, and NULL.
	# string = [ ( leadchar / pair ) [ *( stringchar / pair ) ( trailchar / pair ) ] ]
	#
	# leadchar = LUTF1 / UTFMB
	# LUTF1 = %x01-1F / %x21 / %x24-2A / %x2D-3A / %x3D / %x3F-5B / %x5D-7F
	#
	# trailchar = TUTF1 / UTFMB
	# TUTF1 = %x01-1F / %x21 / %x23-2A / %x2D-3A / %x3D / %x3F-5B / %x5D-7F
	#
	# stringchar = SUTF1 / UTFMB
	# SUTF1 = %x01-21 / %x23-2A / %x2D-3A / %x3D / %x3F-5B / %x5D-7F
	#
	# pair = ESC ( ESC / special / hexpair )
	# special = escaped / SPACE / SHARP / EQUALS
	# escaped = DQUOTE / PLUS / COMMA / SEMI / LANGLE / RANGLE
	#
	# hexstring = SHARP 1*hexpair
	# hexpair = HEX HEX

	@classmethod
	def from_str(cls, schema, expr):
		attribute, escaped_value = expr.split('=', 1)
		if escaped_value.startswith('#'):
			# The "#..." form is used for unknown attribute types and those without
			# an LDAP string encoding. Supporting it would require us to somehow
			# handle the hex-encoded BER encoding of the data. We'll stay away from
			# this mess for now.
			raise ValueError('Hex-encoded RDN assertion values are not supported')
		escaped = False
		hexdigit = None
		encoded_value = b''
		for char in escaped_value:
			if hexdigit is not None:
				encoded_value += bytes.fromhex('%s%s'%(hexdigit, char))
				hexdigit = None
			elif escaped:
				if ord(char) in DN_SPECIAL + (b'\\'[0],):
					encoded_value += char.encode('utf8')
				elif ord(char) in HEXDIGITS:
					hexdigit = char
				else:
					raise ValueError('Invalid escape: \\%s'%char)
				escaped = False
			elif char == '\\':
				escaped = True
			else:
				encoded_value += char.encode('utf8')
		try:
			attribute_type = schema.attribute_types[attribute]
		except KeyError as exc:
			raise ValueError('Invalid RDN attribute type: Attribute type undefined in schema') from exc
		try:
			value = attribute_type.syntax.decode(encoded_value)
		except exceptions.LDAPInvalidAttributeSyntax as exc:
			raise ValueError('Invalid RDN assertion value') from exc
		return cls(schema, attribute, value)

	def __str__(self):
		encoded_value = self.attribute_type.syntax.encode(self.value)
		escaped_value = ''
		for index in range(len(encoded_value)):
			byte = encoded_value[index:index+1]
			if not byte.isascii() or not byte.decode().isprintable():
				escaped_value += '\\%02x'%byte[0]
			elif byte[0] in DN_ESCAPED + (b'\\'[0],):
				escaped_value += '\\' + byte.decode()
			else:
				escaped_value += byte.decode()
		if escaped_value.startswith(' ') or escaped_value.startswith('#'):
			escaped_value = '\\' + escaped_value
		if escaped_value.endswith(' '):
			escaped_value = escaped_value[:-1] + '\\' + escaped_value[-1]
		return '%s=%s'%(self.attribute, escaped_value)

	def __hash__(self):
		return hash(self.attribute_type.oid)

	def __repr__(self):
		return '<ldapserver.dn.RDNAssertion %s>'%str(self)

	def __eq__(self, obj):
		return type(self) is type(obj) and self.attribute_type is obj.attribute_type and \
		       self.attribute_type.equality.match_equal([self.value], obj.value)

	def __ne__(self, obj):
		return not self == obj

	def __setattr__(self, *args):
		raise TypeError('RDNAssertion object is immutable')

	def __delattr__(self, *args):
		raise TypeError('RDNAssertion object is immutable')



import re

from .definitions import MatchingRuleDefinition, MatchingRuleKind
from .. import rfc4518_stringprep, exceptions
from . import syntaxes

class GenericMatchingRuleDefinition(MatchingRuleDefinition):
	def match_equal(self, schema, attribute_values, assertion_value):
		for attribute_value in attribute_values:
			if attribute_value == assertion_value:
				return True
		return False

	def match_less(self, schema, attribute_values, assertion_value):
		for attribute_value in attribute_values:
			if attribute_value < assertion_value:
				return True
		return False

	def match_greater_or_equal(self, schema, attribute_values, assertion_value):
		for attribute_value in attribute_values:
			if attribute_value >= assertion_value:
				return True
		return False

def _substr_match(attribute_value, inital_substring, any_substrings, final_substring):
	if inital_substring:
		if not attribute_value.startswith(inital_substring):
			return False
		attribute_value = attribute_value[len(inital_substring):]
	if final_substring:
		if not attribute_value.endswith(final_substring):
			return False
		attribute_value = attribute_value[:-len(final_substring)]
	for substring in any_substrings:
		index = attribute_value.find(substring)
		if index == -1:
			return False
		attribute_value = attribute_value[index+len(substring):]
	return True

class StringMatchingRuleDefinition(MatchingRuleDefinition):
	def __init__(self, oid, matching_type=rfc4518_stringprep.MatchingType.EXACT_STRING, **kwargs):
		super().__init__(oid, **kwargs)
		self.matching_type = matching_type

	def prepare_assertion_value(self, attribute_value):
		try:
			return rfc4518_stringprep.prepare(attribute_value, self.matching_type)
		except ValueError as exc:
			raise exceptions.LDAPInvalidAttributeSyntax('Assertion value contains characters prohibited by RFC4518') from exc

	def prepare_attribute_value(self, attribute_value):
		try:
			return rfc4518_stringprep.prepare(attribute_value, self.matching_type)
		except ValueError:
			return None

	def match_equal(self, schema, attribute_values, assertion_value):
		assertion_value = self.prepare_assertion_value(assertion_value)
		for attribute_value in attribute_values:
			attribute_value = self.prepare_attribute_value(attribute_value)
			if attribute_value == assertion_value:
				return True
		return False

	def match_less(self, schema, attribute_values, assertion_value):
		assertion_value = self.prepare_assertion_value(assertion_value)
		for attribute_value in attribute_values:
			attribute_value = self.prepare_attribute_value(attribute_value)
			if attribute_value < assertion_value:
				return True
		return False

	def match_greater_or_equal(self, schema, attribute_values, assertion_value):
		assertion_value = self.prepare_assertion_value(assertion_value)
		for attribute_value in attribute_values:
			attribute_value = self.prepare_attribute_value(attribute_value)
			if attribute_value >= assertion_value:
				return True
		return False

	def match_substr(self, schema, attribute_values, inital_substring, any_substrings, final_substring):
		try:
			if inital_substring:
				inital_substring = rfc4518_stringprep.prepare(inital_substring, self.matching_type, rfc4518_stringprep.SubstringType.INITIAL)
			any_substrings = [rfc4518_stringprep.prepare(substring, self.matching_type, rfc4518_stringprep.SubstringType.ANY) for substring in any_substrings]
			if final_substring:
				final_substring = rfc4518_stringprep.prepare(final_substring, self.matching_type, rfc4518_stringprep.SubstringType.FINAL)
		except ValueError as exc:
			raise exceptions.LDAPInvalidAttributeSyntax('Assertion value contains characters prohibited by RFC4518') from exc
		for attribute_value in attribute_values:
			try:
				attribute_value = self.prepare_attribute_value(attribute_value)
				if _substr_match(attribute_value, inital_substring, any_substrings, final_substring):
					return True
			except ValueError:
				pass
		return False

class StringListMatchingRuleDefinition(MatchingRuleDefinition):
	def __init__(self, oid, matching_type=rfc4518_stringprep.MatchingType.EXACT_STRING, **kwargs):
		super().__init__(oid, **kwargs)
		self.matching_type = matching_type

	# Values are both lists of str
	def match_equal(self, schema, attribute_values, assertion_value):
		try:
			assertion_value = [rfc4518_stringprep.prepare(line, self.matching_type) for line in assertion_value]
		except ValueError as exc:
			raise exceptions.LDAPInvalidAttributeSyntax('Assertion value contains characters prohibited by RFC4518') from exc
		for attribute_value in attribute_values:
			try:
				attribute_value = [rfc4518_stringprep.prepare(line, self.matching_type) for line in attribute_value]
				if attribute_value == assertion_value:
					return True
			except ValueError:
				pass
		return False

	def match_substr(self, schema, attribute_values, inital_substring, any_substrings, final_substring):
		try:
			if inital_substring:
				inital_substring = rfc4518_stringprep.prepare(inital_substring, self.matching_type, rfc4518_stringprep.SubstringType.INITIAL)
			any_substrings = [rfc4518_stringprep.prepare(substring, self.matching_type, rfc4518_stringprep.SubstringType.ANY) for substring in any_substrings]
			if final_substring:
				final_substring = rfc4518_stringprep.prepare(final_substring, self.matching_type, rfc4518_stringprep.SubstringType.FINAL)
		except ValueError as exc:
			raise exceptions.LDAPInvalidAttributeSyntax('Assertion value contains characters prohibited by RFC4518') from exc
		for attribute_value in attribute_values:
			try:
				# LF is mapped to SPACE by stringprep, so it is suitable as a seperator
				attribute_value = '\n'.join([rfc4518_stringprep.prepare(line, self.matching_type) for line in attribute_value])
				if _substr_match(attribute_value, inital_substring, any_substrings, final_substring):
					return True
			except ValueError:
				pass
		return False

class FirstComponentMatchingRuleDefinition(MatchingRuleDefinition):
	def __init__(self, oid, attribute_name, matching_rule, compatability_tag=None, **kwargs):
		compatability_tag = compatability_tag or 'FirstComponent:'+matching_rule.compatability_tag
		super().__init__(oid, compatability_tag=compatability_tag, **kwargs)
		self.attribute_name = attribute_name
		self.matching_rule = matching_rule

	def match_equal(self, schema, attribute_values, assertion_value):
		attribute_values = [getattr(value, self.attribute_name)
		                    for value in attribute_values
		                    if hasattr(value, self.attribute_name)]
		return self.matching_rule.match_equal(schema, attribute_values, assertion_value)


class OIDMatchingRuleDefinition(MatchingRuleDefinition):
	NUMERIC_OID_RE = re.compile(r"([0-9]|[1-9][0-9]*)(\.([0-9]|[1-9][0-9]*))*")

	def match_equal(self, schema, attribute_values, assertion_value):
		if not self.NUMERIC_OID_RE.fullmatch(assertion_value):
			assertion_value = schema.get_numeric_oid(assertion_value)
		if assertion_value is None:
			raise exceptions.LDAPInvalidAttributeSyntax('Assertion value is an unknown OID descriptor')
		for attribute_value in attribute_values:
			attribute_value = schema.get_numeric_oid(attribute_value, attribute_value)
			if attribute_value == assertion_value:
				return True
		return False

class StubMatchingRuleDefinition(MatchingRuleDefinition):
	pass

# RFC4517
bitStringMatch = GenericMatchingRuleDefinition('2.5.13.16', name=['bitStringMatch'], syntax=syntaxes.BitString.oid, kind=MatchingRuleKind.EQUALITY)
booleanMatch = GenericMatchingRuleDefinition('2.5.13.13', name=['booleanMatch'], syntax=syntaxes.Boolean.oid, kind=MatchingRuleKind.EQUALITY)
caseExactIA5Match = StringMatchingRuleDefinition('1.3.6.1.4.1.1466.109.114.1', name=['caseExactIA5Match'], syntax=syntaxes.IA5String.oid, matching_type=rfc4518_stringprep.MatchingType.EXACT_STRING, kind=MatchingRuleKind.EQUALITY)
caseExactMatch = StringMatchingRuleDefinition('2.5.13.5', name=['caseExactMatch'], syntax=syntaxes.DirectoryString.oid, matching_type=rfc4518_stringprep.MatchingType.EXACT_STRING, kind=MatchingRuleKind.EQUALITY)
caseExactOrderingMatch = StringMatchingRuleDefinition('2.5.13.6', name=['caseExactOrderingMatch'], syntax=syntaxes.DirectoryString.oid, matching_type=rfc4518_stringprep.MatchingType.EXACT_STRING, kind=MatchingRuleKind.ORDERING)
caseExactSubstringsMatch = StringMatchingRuleDefinition('2.5.13.7', name=['caseExactSubstringsMatch'], syntax=syntaxes.SubstringAssertion.oid, compatability_tag=syntaxes.DirectoryString.oid, matching_type=rfc4518_stringprep.MatchingType.EXACT_STRING, kind=MatchingRuleKind.SUBSTR)
caseIgnoreIA5Match = StringMatchingRuleDefinition('1.3.6.1.4.1.1466.109.114.2', name=['caseIgnoreIA5Match'], syntax=syntaxes.IA5String.oid, matching_type=rfc4518_stringprep.MatchingType.CASE_IGNORE_STRING, kind=MatchingRuleKind.EQUALITY)
caseIgnoreIA5SubstringsMatch = StringMatchingRuleDefinition('1.3.6.1.4.1.1466.109.114.3', name=['caseIgnoreIA5SubstringsMatch'], syntax=syntaxes.SubstringAssertion.oid, compatability_tag=syntaxes.IA5String.oid, matching_type=rfc4518_stringprep.MatchingType.CASE_IGNORE_STRING, kind=MatchingRuleKind.SUBSTR)
caseIgnoreListMatch = StringListMatchingRuleDefinition('2.5.13.11', name=['caseIgnoreListMatch'], syntax=syntaxes.PostalAddress.oid, matching_type=rfc4518_stringprep.MatchingType.CASE_IGNORE_STRING, kind=MatchingRuleKind.EQUALITY)
caseIgnoreListSubstringsMatch = StringListMatchingRuleDefinition('2.5.13.12', name=['caseIgnoreListSubstringsMatch'], syntax=syntaxes.SubstringAssertion.oid, compatability_tag=syntaxes.PostalAddress.oid, matching_type=rfc4518_stringprep.MatchingType.CASE_IGNORE_STRING, kind=MatchingRuleKind.SUBSTR)
caseIgnoreMatch = StringMatchingRuleDefinition('2.5.13.2', name=['caseIgnoreMatch'], syntax=syntaxes.DirectoryString.oid, matching_type=rfc4518_stringprep.MatchingType.CASE_IGNORE_STRING, kind=MatchingRuleKind.EQUALITY)
caseIgnoreOrderingMatch = StringMatchingRuleDefinition('2.5.13.3', name=['caseIgnoreOrderingMatch'], syntax=syntaxes.DirectoryString.oid, matching_type=rfc4518_stringprep.MatchingType.CASE_IGNORE_STRING, kind=MatchingRuleKind.ORDERING)
caseIgnoreSubstringsMatch = StringMatchingRuleDefinition('2.5.13.4', name=['caseIgnoreSubstringsMatch'], syntax=syntaxes.SubstringAssertion.oid, compatability_tag=syntaxes.DirectoryString.oid, matching_type=rfc4518_stringprep.MatchingType.CASE_IGNORE_STRING, kind=MatchingRuleKind.SUBSTR)
directoryStringFirstComponentMatch = FirstComponentMatchingRuleDefinition('2.5.13.31', name=['directoryStringFirstComponentMatch'], syntax=syntaxes.DirectoryString.oid, attribute_name='first_component_string', matching_rule=caseIgnoreMatch, kind=MatchingRuleKind.EQUALITY)
distinguishedNameMatch = GenericMatchingRuleDefinition('2.5.13.1', name=['distinguishedNameMatch'], syntax=syntaxes.DN.oid, kind=MatchingRuleKind.EQUALITY)
generalizedTimeMatch = GenericMatchingRuleDefinition('2.5.13.27', name=['generalizedTimeMatch'], syntax=syntaxes.GeneralizedTime.oid, kind=MatchingRuleKind.EQUALITY)
generalizedTimeOrderingMatch = GenericMatchingRuleDefinition('2.5.13.28', name=['generalizedTimeOrderingMatch'], syntax=syntaxes.GeneralizedTime.oid, kind=MatchingRuleKind.ORDERING)
integerMatch = GenericMatchingRuleDefinition('2.5.13.14', name=['integerMatch'], syntax=syntaxes.INTEGER.oid, kind=MatchingRuleKind.EQUALITY)
integerFirstComponentMatch = FirstComponentMatchingRuleDefinition('2.5.13.29', name=['integerFirstComponentMatch'], syntax=syntaxes.INTEGER.oid, attribute_name='first_component_integer', matching_rule=integerMatch, kind=MatchingRuleKind.EQUALITY)
integerOrderingMatch = GenericMatchingRuleDefinition('2.5.13.15', name=['integerOrderingMatch'], syntax=syntaxes.INTEGER.oid, kind=MatchingRuleKind.ORDERING)
# Optional and implementation-specific, we simply never match
keywordMatch = StubMatchingRuleDefinition('2.5.13.33', name=['keywordMatch'], syntax=syntaxes.DirectoryString.oid, kind=MatchingRuleKind.EQUALITY)
numericStringMatch = StringMatchingRuleDefinition('2.5.13.8', name=['numericStringMatch'], syntax=syntaxes.NumericString.oid, matching_type=rfc4518_stringprep.MatchingType.NUMERIC_STRING, kind=MatchingRuleKind.EQUALITY)
numericStringOrderingMatch = StringMatchingRuleDefinition('2.5.13.9', name=['numericStringOrderingMatch'], syntax=syntaxes.NumericString.oid, matching_type=rfc4518_stringprep.MatchingType.NUMERIC_STRING, kind=MatchingRuleKind.ORDERING)
numericStringSubstringsMatch = StringMatchingRuleDefinition('2.5.13.10', name=['numericStringSubstringsMatch'], syntax=syntaxes.SubstringAssertion.oid, compatability_tag=syntaxes.NumericString.oid, matching_type=rfc4518_stringprep.MatchingType.NUMERIC_STRING, kind=MatchingRuleKind.SUBSTR)
objectIdentifierMatch = OIDMatchingRuleDefinition('2.5.13.0', name=['objectIdentifierMatch'], syntax=syntaxes.OID.oid, kind=MatchingRuleKind.EQUALITY)
objectIdentifierFirstComponentMatch = FirstComponentMatchingRuleDefinition('2.5.13.30', name=['objectIdentifierFirstComponentMatch'], syntax=syntaxes.OID.oid, attribute_name='first_component_oid', matching_rule=objectIdentifierMatch, kind=MatchingRuleKind.EQUALITY)
octetStringMatch = GenericMatchingRuleDefinition('2.5.13.17', name=['octetStringMatch'], syntax=syntaxes.OctetString.oid, kind=MatchingRuleKind.EQUALITY)
octetStringOrderingMatch = GenericMatchingRuleDefinition('2.5.13.18', name=['octetStringOrderingMatch'], syntax=syntaxes.OctetString.oid, kind=MatchingRuleKind.ORDERING)
telephoneNumberMatch = StringMatchingRuleDefinition('2.5.13.20', name=['telephoneNumberMatch'], syntax=syntaxes.TelephoneNumber.oid, matching_type=rfc4518_stringprep.MatchingType.TELEPHONE_NUMBER, kind=MatchingRuleKind.EQUALITY)
telephoneNumberSubstringsMatch = StringMatchingRuleDefinition('2.5.13.21', name=['telephoneNumberSubstringsMatch'], syntax=syntaxes.SubstringAssertion.oid, compatability_tag=syntaxes.TelephoneNumber.oid, matching_type=rfc4518_stringprep.MatchingType.TELEPHONE_NUMBER, kind=MatchingRuleKind.SUBSTR)
uniqueMemberMatch = GenericMatchingRuleDefinition('2.5.13.23', name=['uniqueMemberMatch'], syntax=syntaxes.NameAndOptionalUID.oid, kind=MatchingRuleKind.EQUALITY)
# Optional and implementation-specific, we simply never match
wordMatch = StubMatchingRuleDefinition('2.5.13.32', name=['wordMatch'], syntax=syntaxes.DirectoryString.oid, kind=MatchingRuleKind.EQUALITY)

# RFC4523
certificateExactMatch = StubMatchingRuleDefinition('2.5.13.34', name=['certificateExactMatch'], desc='X.509 Certificate Exact Match', syntax='1.3.6.1.1.15.1', kind=MatchingRuleKind.EQUALITY)
certificateMatch = StubMatchingRuleDefinition('2.5.13.35', name=['certificateMatch'], desc='X.509 Certificate Match', syntax='1.3.6.1.1.15.2', kind=MatchingRuleKind.EQUALITY)
certificatePairExactMatch = StubMatchingRuleDefinition('2.5.13.36', name=['certificatePairExactMatch'], desc='X.509 Certificate Pair Exact Match', syntax='1.3.6.1.1.15.3', kind=MatchingRuleKind.EQUALITY)
certificatePairMatch = StubMatchingRuleDefinition('2.5.13.37', name=['certificatePairMatch'], desc='X.509 Certificate Pair Match', syntax='1.3.6.1.1.15.4', kind=MatchingRuleKind.EQUALITY)
certificateListExactMatch = StubMatchingRuleDefinition('2.5.13.38', name=['certificateListExactMatch'], desc='X.509 Certificate List Exact Match', syntax='1.3.6.1.1.15.5', kind=MatchingRuleKind.EQUALITY)
certificateListMatch = StubMatchingRuleDefinition('2.5.13.39', name=['certificateListMatch'], desc='X.509 Certificate List Match', syntax='1.3.6.1.1.15.6', kind=MatchingRuleKind.EQUALITY)
algorithmIdentifierMatch = StubMatchingRuleDefinition('2.5.13.40', name=['algorithmIdentifierMatch'], desc='X.509 Algorithm Identifier Match', syntax='1.3.6.1.1.15.7', kind=MatchingRuleKind.EQUALITY)

# RFC3112
authPasswordExactMatch = StubMatchingRuleDefinition('1.3.6.1.4.1.4203.1.2.2', name=['authPasswordExactMatch'], desc='authentication password exact matching rule', syntax='1.3.6.1.4.1.1466.115.121.1.40', kind=MatchingRuleKind.EQUALITY)
authPasswordMatch = StubMatchingRuleDefinition('1.3.6.1.4.1.4203.1.2.3', name=['authPasswordMatch'], desc='authentication password matching rule', syntax='1.3.6.1.4.1.1466.115.121.1.40', kind=MatchingRuleKind.EQUALITY)

ALL = (
	# RFC4517
	bitStringMatch,
	booleanMatch,
	caseExactIA5Match,
	caseExactMatch,
	caseExactOrderingMatch,
	caseExactSubstringsMatch,
	caseIgnoreIA5Match,
	caseIgnoreIA5SubstringsMatch,
	caseIgnoreListMatch,
	caseIgnoreListSubstringsMatch,
	caseIgnoreMatch,
	caseIgnoreOrderingMatch,
	caseIgnoreSubstringsMatch,
	directoryStringFirstComponentMatch,
	distinguishedNameMatch,
	generalizedTimeMatch,
	generalizedTimeOrderingMatch,
	integerFirstComponentMatch,
	integerMatch,
	integerOrderingMatch,
	keywordMatch,
	numericStringMatch,
	numericStringOrderingMatch,
	numericStringSubstringsMatch,
	objectIdentifierFirstComponentMatch,
	objectIdentifierMatch,
	octetStringMatch,
	octetStringOrderingMatch,
	telephoneNumberMatch,
	telephoneNumberSubstringsMatch,
	uniqueMemberMatch,
	wordMatch,

	# RFC4523
	certificateExactMatch,
	certificateMatch,
	certificatePairExactMatch,
	certificatePairMatch,
	certificateListExactMatch,
	certificateListMatch,
	algorithmIdentifierMatch,

	# RFC3112
	authPasswordExactMatch,
	authPasswordMatch,
)



import unicodedata
import stringprep
import enum

SPACE = 0x0020

MAPPED_TO_NOTHING = {
	# SOFT HYPHEN (U+00AD)
	0x00AD,
	# MONGOLIAN TODO SOFT HYPHEN (U+1806)
	0x1806,
	# COMBINING GRAPHEME JOINER (U+034F)
	0x034F,
	# VARIATION SELECTORs (U+180B-180D, FF00-FE0F)
	0x180B, 0x180C, 0x180D,
	0xFE00, 0xFE01, 0xFE02, 0xFE03, 0xFE04, 0xFE05, 0xFE06, 0xFE07, 0xFE08,
	0xFE09, 0xFE0A, 0xFE0B, 0xFE0C, 0xFE0D, 0xFE0E, 0xFE0F,
	# OBJECT REPLACEMENT CHARACTER (U+FFFC)
	0xFFFC,
	# ZERO WIDTH SPACE (U+200B)
	0x200B,
}

MAPPED_TO_SPACE = {
	# CHARACTER TABULATION (U+0009)
	0x0009,
	# LINE FEED (LF) (U+000A)
	0x000A,
	# LINE TABULATION (U+000B)
	0x000B,
	# FORM FEED (FF) (U+000C)
	0x000C,
	# CARRIAGE RETURN (CR) (U+000D)
	0x000D,
	# NEXT LINE (NEL) (U+0085)
	0x0085,
	# All other control code (e.g., Cc) points or code points with a
	# control function (e.g., Cf)
	0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007, 0x0008,
	0x000E, 0x000F, 0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016,
	0x0017, 0x0018, 0x0019, 0x001A, 0x001B, 0x001C, 0x001D, 0x001E, 0x001F,
	0x007F, 0x0080, 0x0081, 0x0082, 0x0083, 0x0084,
	0x0086, 0x0087, 0x0088, 0x0089, 0x008A, 0x008B, 0x008C, 0x008D, 0x008E,
	0x008F, 0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097,
	0x0098, 0x0099, 0x009A, 0x009B, 0x009C, 0x009D, 0x009E, 0x009F,
	0x06DD,
	0x070F,
	0x180E,
	0x200C, 0x200D, 0x200E, 0x200F,
	0x202A, 0x202B, 0x202C, 0x202D, 0x202E,
	0x2060, 0x2061, 0x2062, 0x2063,
	0x206A, 0x206B, 0x206C, 0x206D, 0x206E, 0x206F,
	0xFEFF,
	0xFFF9, 0xFFFA, 0xFFFB,
	0x1D173, 0x1D174, 0x1D175, 0x1D176, 0x1D177, 0x1D178, 0x1D179, 0x1D17A,
	0xE0001,
	0xE0020, 0xE0021, 0xE0022, 0xE0023, 0xE0024, 0xE0025, 0xE0026, 0xE0027,
	0xE0028, 0xE0029, 0xE002A, 0xE002B, 0xE002C, 0xE002D, 0xE002E, 0xE002F,
	0xE0030, 0xE0031, 0xE0032, 0xE0033, 0xE0034, 0xE0035, 0xE0036, 0xE0037,
	0xE0038, 0xE0039, 0xE003A, 0xE003B, 0xE003C, 0xE003D, 0xE003E, 0xE003F,
	0xE0040, 0xE0041, 0xE0042, 0xE0043, 0xE0044, 0xE0045, 0xE0046, 0xE0047,
	0xE0048, 0xE0049, 0xE004A, 0xE004B, 0xE004C, 0xE004D, 0xE004E, 0xE004F,
	0xE0050, 0xE0051, 0xE0052, 0xE0053, 0xE0054, 0xE0055, 0xE0056, 0xE0057,
	0xE0058, 0xE0059, 0xE005A, 0xE005B, 0xE005C, 0xE005D, 0xE005E, 0xE005F,
	0xE0060, 0xE0061, 0xE0062, 0xE0063, 0xE0064, 0xE0065, 0xE0066, 0xE0067,
	0xE0068, 0xE0069, 0xE006A, 0xE006B, 0xE006C, 0xE006D, 0xE006E, 0xE006F,
	0xE0070, 0xE0071, 0xE0072, 0xE0073, 0xE0074, 0xE0075, 0xE0076, 0xE0077,
	0xE0078, 0xE0079, 0xE007A, 0xE007B, 0xE007C, 0xE007D, 0xE007E, 0xE007F,
}

# unicodedata.combining does not seem to reflect this table and the RFC
# says the embedded table is definitive.
COMBINING_MARKS = {
	0x0300, 0x0301, 0x0302, 0x0303, 0x0304, 0x0305, 0x0306, 0x0307, 0x0308,
	0x0309, 0x030A, 0x030B, 0x030C, 0x030D, 0x030E, 0x030F, 0x0310, 0x0311,
	0x0312, 0x0313, 0x0314, 0x0315, 0x0316, 0x0317, 0x0318, 0x0319, 0x031A,
	0x031B, 0x031C, 0x031D, 0x031E, 0x031F, 0x0320, 0x0321, 0x0322, 0x0323,
	0x0324, 0x0325, 0x0326, 0x0327, 0x0328, 0x0329, 0x032A, 0x032B, 0x032C,
	0x032D, 0x032E, 0x032F, 0x0330, 0x0331, 0x0332, 0x0333, 0x0334, 0x0335,
	0x0336, 0x0337, 0x0338, 0x0339, 0x033A, 0x033B, 0x033C, 0x033D, 0x033E,
	0x033F, 0x0340, 0x0341, 0x0342, 0x0343, 0x0344, 0x0345, 0x0346, 0x0347,
	0x0348, 0x0349, 0x034A, 0x034B, 0x034C, 0x034D, 0x034E, 0x034F, 0x0360,
	0x0361, 0x0362, 0x0363, 0x0364, 0x0365, 0x0366, 0x0367, 0x0368, 0x0369,
	0x036A, 0x036B, 0x036C, 0x036D, 0x036E, 0x036F, 0x0483, 0x0484, 0x0485,
	0x0486, 0x0488, 0x0489, 0x0591, 0x0592, 0x0593, 0x0594, 0x0595, 0x0596,
	0x0597, 0x0598, 0x0599, 0x059A, 0x059B, 0x059C, 0x059D, 0x059E, 0x059F,
	0x05A0, 0x05A1, 0x05A3, 0x05A4, 0x05A5, 0x05A6, 0x05A7, 0x05A8, 0x05A9,
	0x05AA, 0x05AB, 0x05AC, 0x05AD, 0x05AE, 0x05AF, 0x05B0, 0x05B1, 0x05B2,
	0x05B3, 0x05B4, 0x05B5, 0x05B6, 0x05B7, 0x05B8, 0x05B9, 0x05BB, 0x05BC,
	0x05BF, 0x05C1, 0x05C2, 0x05C4, 0x064B, 0x064C, 0x064D, 0x064E, 0x064F,
	0x0650, 0x0651, 0x0652, 0x0653, 0x0654, 0x0655, 0x0670, 0x06D6, 0x06D7,
	0x06D8, 0x06D9, 0x06DA, 0x06DB, 0x06DC, 0x06DE, 0x06DF, 0x06E0, 0x06E1,
	0x06E2, 0x06E3, 0x06E4, 0x06E7, 0x06E8, 0x06EA, 0x06EB, 0x06EC, 0x06ED,
	0x0711, 0x0730, 0x0731, 0x0732, 0x0733, 0x0734, 0x0735, 0x0736, 0x0737,
	0x0738, 0x0739, 0x073A, 0x073B, 0x073C, 0x073D, 0x073E, 0x073F, 0x0740,
	0x0741, 0x0742, 0x0743, 0x0744, 0x0745, 0x0746, 0x0747, 0x0748, 0x0749,
	0x074A, 0x07A6, 0x07A7, 0x07A8, 0x07A9, 0x07AA, 0x07AB, 0x07AC, 0x07AD,
	0x07AE, 0x07AF, 0x07B0, 0x0901, 0x0902, 0x0903, 0x093C, 0x093E, 0x093F,
	0x0940, 0x0941, 0x0942, 0x0943, 0x0944, 0x0945, 0x0946, 0x0947, 0x0948,
	0x0949, 0x094A, 0x094B, 0x094C, 0x094D, 0x094E, 0x094F, 0x0951, 0x0952,
	0x0953, 0x0954, 0x0962, 0x0963, 0x0981, 0x0982, 0x0983, 0x09BC, 0x09BE,
	0x09BF, 0x09C0, 0x09C1, 0x09C2, 0x09C3, 0x09C4, 0x09C7, 0x09C8, 0x09CB,
	0x09CC, 0x09CD, 0x09D7, 0x09E2, 0x09E3, 0x0A02, 0x0A3C, 0x0A3E, 0x0A3F,
	0x0A40, 0x0A41, 0x0A42, 0x0A47, 0x0A48, 0x0A4B, 0x0A4C, 0x0A4D, 0x0A70,
	0x0A71, 0x0A81, 0x0A82, 0x0A83, 0x0ABC, 0x0ABE, 0x0ABF, 0x0AC0, 0x0AC1,
	0x0AC2, 0x0AC3, 0x0AC4, 0x0AC5, 0x0AC7, 0x0AC8, 0x0AC9, 0x0ACB, 0x0ACC,
	0x0ACD, 0x0B01, 0x0B02, 0x0B03, 0x0B3C, 0x0B3E, 0x0B3F, 0x0B40, 0x0B41,
	0x0B42, 0x0B43, 0x0B47, 0x0B48, 0x0B4B, 0x0B4C, 0x0B4D, 0x0B56, 0x0B57,
	0x0B82, 0x0BBE, 0x0BBF, 0x0BC0, 0x0BC1, 0x0BC2, 0x0BC6, 0x0BC7, 0x0BC8,
	0x0BCA, 0x0BCB, 0x0BCC, 0x0BCD, 0x0BD7, 0x0C01, 0x0C02, 0x0C03, 0x0C3E,
	0x0C3F, 0x0C40, 0x0C41, 0x0C42, 0x0C43, 0x0C44, 0x0C46, 0x0C47, 0x0C48,
	0x0C4A, 0x0C4B, 0x0C4C, 0x0C4D, 0x0C55, 0x0C56, 0x0C82, 0x0C83, 0x0CBE,
	0x0CBF, 0x0CC0, 0x0CC1, 0x0CC2, 0x0CC3, 0x0CC4, 0x0CC6, 0x0CC7, 0x0CC8,
	0x0CCA, 0x0CCB, 0x0CCC, 0x0CCD, 0x0CD5, 0x0CD6, 0x0D02, 0x0D03, 0x0D3E,
	0x0D3F, 0x0D40, 0x0D41, 0x0D42, 0x0D43, 0x0D46, 0x0D47, 0x0D48, 0x0D4A,
	0x0D4B, 0x0D4C, 0x0D4D, 0x0D57, 0x0D82, 0x0D83, 0x0DCA, 0x0DCF, 0x0DD0,
	0x0DD1, 0x0DD2, 0x0DD3, 0x0DD4, 0x0DD6, 0x0DD8, 0x0DD9, 0x0DDA, 0x0DDB,
	0x0DDC, 0x0DDD, 0x0DDE, 0x0DDF, 0x0DF2, 0x0DF3, 0x0E31, 0x0E34, 0x0E35,
	0x0E36, 0x0E37, 0x0E38, 0x0E39, 0x0E3A, 0x0E47, 0x0E48, 0x0E49, 0x0E4A,
	0x0E4B, 0x0E4C, 0x0E4D, 0x0E4E, 0x0EB1, 0x0EB4, 0x0EB5, 0x0EB6, 0x0EB7,
	0x0EB8, 0x0EB9, 0x0EBB, 0x0EBC, 0x0EC8, 0x0EC9, 0x0ECA, 0x0ECB, 0x0ECC,
	0x0ECD, 0x0F18, 0x0F19, 0x0F35, 0x0F37, 0x0F39, 0x0F3E, 0x0F3F, 0x0F71,
	0x0F72, 0x0F73, 0x0F74, 0x0F75, 0x0F76, 0x0F77, 0x0F78, 0x0F79, 0x0F7A,
	0x0F7B, 0x0F7C, 0x0F7D, 0x0F7E, 0x0F7F, 0x0F80, 0x0F81, 0x0F82, 0x0F83,
	0x0F84, 0x0F86, 0x0F87, 0x0F90, 0x0F91, 0x0F92, 0x0F93, 0x0F94, 0x0F95,
	0x0F96, 0x0F97, 0x0F99, 0x0F9A, 0x0F9B, 0x0F9C, 0x0F9D, 0x0F9E, 0x0F9F,
	0x0FA0, 0x0FA1, 0x0FA2, 0x0FA3, 0x0FA4, 0x0FA5, 0x0FA6, 0x0FA7, 0x0FA8,
	0x0FA9, 0x0FAA, 0x0FAB, 0x0FAC, 0x0FAD, 0x0FAE, 0x0FAF, 0x0FB0, 0x0FB1,
	0x0FB2, 0x0FB3, 0x0FB4, 0x0FB5, 0x0FB6, 0x0FB7, 0x0FB8, 0x0FB9, 0x0FBA,
	0x0FBB, 0x0FBC, 0x0FC6, 0x102C, 0x102D, 0x102E, 0x102F, 0x1030, 0x1031,
	0x1032, 0x1036, 0x1037, 0x1038, 0x1039, 0x1056, 0x1057, 0x1058, 0x1059,
	0x1712, 0x1713, 0x1714, 0x1732, 0x1733, 0x1734, 0x1752, 0x1753, 0x1772,
	0x1773, 0x17B4, 0x17B5, 0x17B6, 0x17B7, 0x17B8, 0x17B9, 0x17BA, 0x17BB,
	0x17BC, 0x17BD, 0x17BE, 0x17BF, 0x17C0, 0x17C1, 0x17C2, 0x17C3, 0x17C4,
	0x17C5, 0x17C6, 0x17C7, 0x17C8, 0x17C9, 0x17CA, 0x17CB, 0x17CC, 0x17CD,
	0x17CE, 0x17CF, 0x17D0, 0x17D1, 0x17D2, 0x17D3, 0x180B, 0x180C, 0x180D,
	0x18A9, 0x20D0, 0x20D1, 0x20D2, 0x20D3, 0x20D4, 0x20D5, 0x20D6, 0x20D7,
	0x20D8, 0x20D9, 0x20DA, 0x20DB, 0x20DC, 0x20DD, 0x20DE, 0x20DF, 0x20E0,
	0x20E1, 0x20E2, 0x20E3, 0x20E4, 0x20E5, 0x20E6, 0x20E7, 0x20E8, 0x20E9,
	0x20EA, 0x302A, 0x302B, 0x302C, 0x302D, 0x302E, 0x302F, 0x3099, 0x309A,
	0xFB1E, 0xFE00, 0xFE01, 0xFE02, 0xFE03, 0xFE04, 0xFE05, 0xFE06, 0xFE07,
	0xFE08, 0xFE09, 0xFE0A, 0xFE0B, 0xFE0C, 0xFE0D, 0xFE0E, 0xFE0F, 0xFE20,
	0xFE21, 0xFE22, 0xFE23, 0x1D165, 0x1D166, 0x1D167, 0x1D168, 0x1D169,
	0x1D16D, 0x1D16E, 0x1D16F, 0x1D170, 0x1D171, 0x1D172, 0x1D17B, 0x1D17C,
	0x1D17D, 0x1D17E, 0x1D17F, 0x1D180, 0x1D181, 0x1D182, 0x1D185, 0x1D186,
	0x1D187, 0x1D188, 0x1D189, 0x1D18A, 0x1D18B, 0x1D1AA, 0x1D1AB, 0x1D1AC,
	0x1D1AD,
}

class MatchingType(enum.Enum):
	CASE_IGNORE_STRING = enum.auto()
	EXACT_STRING = enum.auto()
	NUMERIC_STRING = enum.auto()
	TELEPHONE_NUMBER = enum.auto()

class SubstringType(enum.Enum):
	NONE = enum.auto()
	INITIAL = enum.auto()
	ANY = enum.auto()
	FINAL = enum.auto()

def prepare(value, matching_type=MatchingType.EXACT_STRING,
            substring_type=SubstringType.NONE):
	# Algortihm according to RFC 4518
	#
	# 1) Transcode: value is already a Unicode string, no transcoding needed
	# 2) Map
	value = prepare_map(value, matching_type=matching_type)
	# 3) Normalize
	value = prepare_normalize(value)
	# 4) Prohibit
	if prepare_check_prohibited(value):
		raise ValueError('Stringprep "prohibit" stage rejected input')
	# 5) Check bidi: Bidirectional characters are ignored.
	# 6) Insignificant Character Handling
	value = prepare_insignificant_characters(value, matching_type=matching_type,
	                                         substring_type=substring_type)
	return value

def prepare_map(value, matching_type=MatchingType.EXACT_STRING):
	# 2.2.  Map
	#
	# SOFT HYPHEN (U+00AD) and MONGOLIAN TODO SOFT HYPHEN (U+1806) code
	# points are mapped to nothing.  COMBINING GRAPHEME JOINER (U+034F) and
	# VARIATION SELECTORs (U+180B-180D, FF00-FE0F) code points are also
	# mapped to nothing.  The OBJECT REPLACEMENT CHARACTER (U+FFFC) is
	# mapped to nothing.
	#
	# CHARACTER TABULATION (U+0009), LINE FEED (LF) (U+000A), LINE
	# TABULATION (U+000B), FORM FEED (FF) (U+000C), CARRIAGE RETURN (CR)
	# (U+000D), and NEXT LINE (NEL) (U+0085) are mapped to SPACE (U+0020).
	#
	# All other control code (e.g., Cc) points or code points with a
	# control function (e.g., Cf) are mapped to nothing.  The following is
	# a complete list of these code points: U+0000-0008, 000E-001F, 007F-
	# 0084, 0086-009F, 06DD, 070F, 180E, 200C-200F, 202A-202E, 2060-2063,
	# 206A-206F, FEFF, FFF9-FFFB, 1D173-1D17A, E0001, E0020-E007F.
	#
	# ZERO WIDTH SPACE (U+200B) is mapped to nothing.  All other code
	# points with Separator (space, line, or paragraph) property (e.g., Zs,
	# Zl, or Zp) are mapped to SPACE (U+0020).  The following is a complete
	# list of these code points: U+0020, 00A0, 1680, 2000-200A, 2028-2029,
	# 202F, 205F, 3000.
	#
	# For case ignore, numeric, and stored prefix string matching rules,
	# characters are case folded per B.2 of [RFC3454].
	#
	# The output is the mapped string.
	new_value = ''
	for char in value:
		if ord(char) in MAPPED_TO_NOTHING:
			continue
		if ord(char) in MAPPED_TO_SPACE:
			char = ' '
		# No idea what "stored prefix string matching" is supposed to be
		if matching_type in (MatchingType.CASE_IGNORE_STRING,
		                     MatchingType.NUMERIC_STRING):
			char = stringprep.map_table_b2(char)
		new_value += char
	return new_value

def prepare_normalize(value):
	# 2.3.  Normalize
	#
	# The input string is to be normalized to Unicode Form KC
	# (compatibility composed) as described in [UAX15].  The output is the
	# normalized string.
	return unicodedata.normalize('NFKC', value)

def prepare_check_prohibited(value):
	# 2.4.  Prohibit
	#
	# All Unassigned code points are prohibited.  Unassigned code points
	# are listed in Table A.1 of [RFC3454].
	#
	# Characters that, per Section 5.8 of [RFC3454], change display
	# properties or are deprecated are prohibited.  These characters are
	# listed in Table C.8 of [RFC3454].
	#
	# Private Use code points are prohibited.  These characters are listed
	# in Table C.3 of [RFC3454].
	#
	# All non-character code points are prohibited.  These code points are
	# listed in Table C.4 of [RFC3454].
	#
	# Surrogate codes are prohibited.  These characters are listed in Table
	# C.5 of [RFC3454].
	#
	# The REPLACEMENT CHARACTER (U+FFFD) code point is prohibited.
	#
	# The step fails if the input string contains any prohibited code
	# point.  Otherwise, the output is the input string.
	for char in value:
		# pylint: disable=too-many-boolean-expressions
		if stringprep.in_table_a1(char) or \
		   stringprep.in_table_c8(char) or \
		   stringprep.in_table_c3(char) or \
		   stringprep.in_table_c4(char) or \
		   stringprep.in_table_c5(char) or \
		   ord(char) == 0xFFFD:
			return True
	return False

def prepare_insignificant_characters(value,
                                     matching_type=MatchingType.EXACT_STRING,
                                     substring_type=SubstringType.NONE):
	# 2.6.  Insignificant Character Handling
	#
	# In this step, the string is modified to ensure proper handling of
	# characters insignificant to the matching rule.  This modification
	# differs from matching rule to matching rule.
	#
	# Section 2.6.1 applies to case ignore and exact string matching.
	# Section 2.6.2 applies to numericString matching.
	# Section 2.6.3 applies to telephoneNumber matching.
	if matching_type in (MatchingType.CASE_IGNORE_STRING,
	                     MatchingType.EXACT_STRING):
		return prepare_insignificant_space(value, substring_type=substring_type)
	if matching_type == MatchingType.NUMERIC_STRING:
		return prepare_insignificant_numeric_string(value)
	if matching_type == MatchingType.TELEPHONE_NUMBER:
		return prepare_insignificant_telephone_number(value)
	raise ValueError('Invalid matching type')

def prepare_insignificant_space(value, substring_type=SubstringType.NONE):
	# 2.6.1.  Insignificant Space Handling
	#
	# For the purposes of this section, a space is defined to be the SPACE
	# (U+0020) code point followed by no combining marks.
	#
	#     NOTE - The previous steps ensure that the string cannot contain
	#            any code points in the separator class, other than SPACE
	#            (U+0020).
	#
	# For input strings that are attribute values or non-substring
	# assertion values:  If the input string contains no non-space
	# character, then the output is exactly two SPACEs.  Otherwise (the
	# input string contains at least one non-space character), the string
	# is modified such that the string starts with exactly one space
	# character, ends with exactly one SPACE character, and any inner
	# (non-empty) sequence of space characters is replaced with exactly two
	# SPACE characters.  For instance, the input strings
	# "foo<SPACE>bar<SPACE><SPACE>", result in the output
	# "<SPACE>foo<SPACE><SPACE>bar<SPACE>".
	#
	# For input strings that are substring assertion values: If the string
	# being prepared contains no non-space characters, then the output
	# string is exactly one SPACE.  Otherwise, the following steps are
	# taken:
	#
	# - Any inner (non-empty) sequence of space characters is replaced
  # with exactly two SPACE characters; [ERRATA 1757]
	#
	# -  If the input string is an initial substring, it is modified to
	# start with exactly one SPACE character;
	#
	# -  If the input string is an initial or an any substring that ends in
	# one or more space characters, it is modified to end with exactly
	# one SPACE character;
	#
	# -  If the input string is an any or a final substring that starts in
	# one or more space characters, it is modified to start with exactly
	# one SPACE character; and
	#
	# -  If the input string is a final substring, it is modified to end
	# with exactly one SPACE character.
	#
	# For instance, for the input string "foo<SPACE>bar<SPACE><SPACE>" as
	# an initial substring, the output would be
	# "<SPACE>foo<SPACE><SPACE>bar<SPACE>".  As an any or final substring,
	# the same input would result in "foo<SPACE><SPACE>bar<SPACE>".
	# [ERRATA 1758]

	# First we replace all spaces followed by no combining mark with U+FFFD.
	# U+FFFD is used because is it one of the prohibited characters.
	# Additionally we collapse all sequences of one or more SPACEs into
	# exactly two SPACEs.
	PLACEHOLDER = '\uFFFD' # pylint: disable=invalid-name
	new_value = ''
	for i, char in enumerate(value):
		if ord(char) != SPACE:
			new_value += char
		elif i + 1 < len(value) and ord(value[i + 1]) in COMBINING_MARKS:
			new_value += ' '
		elif not new_value or new_value[-1] != PLACEHOLDER:
			new_value += 2*PLACEHOLDER
	value = new_value
	if substring_type == SubstringType.NONE:
		value = PLACEHOLDER + value.strip(PLACEHOLDER) + PLACEHOLDER
	else:
		if substring_type == SubstringType.INITIAL:
			value = PLACEHOLDER + value.lstrip(PLACEHOLDER)
		if substring_type in (SubstringType.INITIAL, SubstringType.ANY) and value.endswith(PLACEHOLDER):
			value = value.rstrip(PLACEHOLDER) + PLACEHOLDER
		if substring_type in (SubstringType.ANY, SubstringType.FINAL) and value.startswith(PLACEHOLDER):
			value = PLACEHOLDER + value.lstrip(PLACEHOLDER)
		if substring_type == SubstringType.FINAL:
			value = value.rstrip(PLACEHOLDER) + PLACEHOLDER
	return value.replace(PLACEHOLDER, ' ')

def prepare_insignificant_numeric_string(value):
	# 2.6.2.  numericString Insignificant Character Handling
	#
	# For the purposes of this section, a space is defined to be the SPACE
	# (U+0020) code point followed by no combining marks.
	#
	# All spaces are regarded as insignificant and are to be removed.
	#
	# For example, removal of spaces from the Form KC string:
	#     "<SPACE><SPACE>123<SPACE><SPACE>456<SPACE><SPACE>"
	# would result in the output string:
	#     "123456"
	# and the Form KC string:
	#     "<SPACE><SPACE><SPACE>"
	# would result in the output string:
	#     "" (an empty string).
	new_value = ''
	# pylint: disable=consider-using-enumerate
	for i in range(len(value)):
		if ord(value[i]) == SPACE:
			if i + 1 >= len(value) or ord(value[i + 1]) not in COMBINING_MARKS:
				continue
		new_value += value[i]
	return new_value

def prepare_insignificant_telephone_number(value):
	# 2.6.3.  telephoneNumber Insignificant Character Handling
	#
	# For the purposes of this section, a hyphen is defined to be a
	# HYPHEN-MINUS (U+002D), ARMENIAN HYPHEN (U+058A), HYPHEN (U+2010),
	# NON-BREAKING HYPHEN (U+2011), MINUS SIGN (U+2212), SMALL HYPHEN-MINUS
	# (U+FE63), or FULLWIDTH HYPHEN-MINUS (U+FF0D) code point followed by
	# no combining marks and a space is defined to be the SPACE (U+0020)
	# code point followed by no combining marks.
	#
	# All hyphens and spaces are considered insignificant and are to be
	# removed.
	#
	# For example, removal of hyphens and spaces from the Form KC string:
	#     "<SPACE><HYPHEN>123<SPACE><SPACE>456<SPACE><HYPHEN>"
	# would result in the output string:
	#     "123456"
	# and the Form KC string:
	#     "<HYPHEN><HYPHEN><HYPHEN>"
	# would result in the (empty) output string:
	#     "".
	hyphen_chars = {0x002D, 0x058A, 0x2010, 0x2011, 0x2212, 0xFE63, 0xFF0D}
	new_value = ''
	# pylint: disable=consider-using-enumerate
	for i in range(len(value)):
		if ord(value[i]) == SPACE or ord(value[i]) in hyphen_chars:
			if i + 1 >= len(value) or ord(value[i + 1]) not in COMBINING_MARKS:
				continue
		new_value += value[i]
	return new_value



import os
import setuptools

with open('README.md', 'r', encoding='utf-8') as f:
	long_description = f.read()

setuptools.setup(
	name='ldapserver',
	version=os.environ.get('PACKAGE_VERSION', 'local'),
	author='Julian Rother',
	author_email='python-ldapserver@jrother.eu',
	description='Library to implement special-purpose LDAP servers',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://git.cccv.de/uffd/python-ldapserver',
	classifiers=[
		'Programming Language :: Python :: 3',
		'Development Status :: 2 - Pre-Alpha',
		'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
		'Operating System :: OS Independent',
		'Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP',
	],
	package_dir={'': '.'},
	packages=setuptools.find_packages(where='.'),
	python_requires='>=3.9',
)



from .ldap import LDAPResultCode

class LDAPError(Exception):
	'''Base class for all LDAP errors'''

	RESULT_CODE: LDAPResultCode

	def __init__(self, message=''):
		super().__init__()
		self.code = self.RESULT_CODE
		self.message = message

#class LDAPSuccess(LDAPError):
#	RESULT_CODE = LDAPResultCode.success

class LDAPOperationsError(LDAPError):
	RESULT_CODE = LDAPResultCode.operationsError

class LDAPProtocolError(LDAPError):
	RESULT_CODE = LDAPResultCode.protocolError

class LDAPTimeLimitExceeded(LDAPError):
	RESULT_CODE = LDAPResultCode.timeLimitExceeded

class LDAPSizeLimitExceeded(LDAPError):
	RESULT_CODE = LDAPResultCode.sizeLimitExceeded

#class LDAPCompareFalse(LDAPError):
#	RESULT_CODE = LDAPResultCode.compareFalse

#class LDAPCompareTrue(LDAPError):
#	RESULT_CODE = LDAPResultCode.compareTrue

class LDAPAuthMethodNotSupported(LDAPError):
	RESULT_CODE = LDAPResultCode.authMethodNotSupported

class LDAPStrongerAuthRequired(LDAPError):
	RESULT_CODE = LDAPResultCode.strongerAuthRequired

#class LDAPReferral(LDAPError):
#	RESULT_CODE = LDAPResultCode.referral

class LDAPAdminLimitExceeded(LDAPError):
	RESULT_CODE = LDAPResultCode.adminLimitExceeded

class LDAPUnavailableCriticalExtension(LDAPError):
	RESULT_CODE = LDAPResultCode.unavailableCriticalExtension

class LDAPConfidentialityRequired(LDAPError):
	RESULT_CODE = LDAPResultCode.confidentialityRequired

#class LDAPSaslBindInProgress(LDAPError):
#	RESULT_CODE = LDAPResultCode.saslBindInProgress

class LDAPNoSuchAttribute(LDAPError):
	RESULT_CODE = LDAPResultCode.noSuchAttribute

class LDAPUndefinedAttributeType(LDAPError):
	RESULT_CODE = LDAPResultCode.undefinedAttributeType

class LDAPInappropriateMatching(LDAPError):
	RESULT_CODE = LDAPResultCode.inappropriateMatching

class LDAPConstraintViolation(LDAPError):
	RESULT_CODE = LDAPResultCode.constraintViolation

class LDAPAttributeOrValueExists(LDAPError):
	RESULT_CODE = LDAPResultCode.attributeOrValueExists

class LDAPInvalidAttributeSyntax(LDAPError):
	RESULT_CODE = LDAPResultCode.invalidAttributeSyntax

class LDAPNoSuchObject(LDAPError):
	RESULT_CODE = LDAPResultCode.noSuchObject

class LDAPAliasProblem(LDAPError):
	RESULT_CODE = LDAPResultCode.aliasProblem

class LDAPInvalidDNSyntax(LDAPError):
	RESULT_CODE = LDAPResultCode.invalidDNSyntax

class LDAPAliasDereferencingProblem(LDAPError):
	RESULT_CODE = LDAPResultCode.aliasDereferencingProblem

class LDAPInappropriateAuthentication(LDAPError):
	RESULT_CODE = LDAPResultCode.inappropriateAuthentication

class LDAPInvalidCredentials(LDAPError):
	RESULT_CODE = LDAPResultCode.invalidCredentials

class LDAPInsufficientAccessRights(LDAPError):
	RESULT_CODE = LDAPResultCode.insufficientAccessRights

class LDAPBusy(LDAPError):
	RESULT_CODE = LDAPResultCode.busy

class LDAPUnavailable(LDAPError):
	RESULT_CODE = LDAPResultCode.unavailable

class LDAPUnwillingToPerform(LDAPError):
	RESULT_CODE = LDAPResultCode.unwillingToPerform

class LDAPLoopDetect(LDAPError):
	RESULT_CODE = LDAPResultCode.loopDetect

class LDAPNamingViolation(LDAPError):
	RESULT_CODE = LDAPResultCode.namingViolation

class LDAPObjectClassViolation(LDAPError):
	RESULT_CODE = LDAPResultCode.objectClassViolation

class LDAPNotAllowedOnNonLeaf(LDAPError):
	RESULT_CODE = LDAPResultCode.notAllowedOnNonLeaf

class LDAPNotAllowedOnRDN(LDAPError):
	RESULT_CODE = LDAPResultCode.notAllowedOnRDN

class LDAPEntryAlreadyExists(LDAPError):
	RESULT_CODE = LDAPResultCode.entryAlreadyExists

class LDAPObjectClassModsProhibited(LDAPError):
	RESULT_CODE = LDAPResultCode.objectClassModsProhibited

class LDAPAffectsMultipleDSAs(LDAPError):
	RESULT_CODE = LDAPResultCode.affectsMultipleDSAs

class LDAPOther(LDAPError):
	RESULT_CODE = LDAPResultCode.other
